'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

exports.default = function (_ref) {
    var t = _ref.types;

    /**
     * If the decorator expressions are non-identifiers, hoist them to before the class so we can be sure
     * that they are evaluated in order.
     */
    function applyEnsureOrdering(path) {
        // TODO: This should probably also hoist computed properties.
        var decorators = (path.isClass() ? [path].concat(path.get('body.body')) : path.get('properties')).reduce(function (acc, prop) {
            return acc.concat(prop.node.decorators || []);
        }, []);

        var identDecorators = decorators.filter(function (decorator) {
            return !t.isIdentifier(decorator.expression);
        });
        if (identDecorators.length === 0) return;

        return t.sequenceExpression(identDecorators.map(function (decorator) {
            var expression = decorator.expression;
            var id = decorator.expression = path.scope.generateDeclaredUidIdentifier('dec');
            return t.assignmentExpression('=', id, expression);
        }).concat([path.node]));
    }

    /**
     * Given a class expression with class-level decorators, create a new expression
     * with the proper decorated behavior.
     */
    function applyClassDecorators(classPath) {
        var decorators = classPath.node.decorators || [];
        classPath.node.decorators = null;

        if (decorators.length === 0) return;

        var name = classPath.scope.generateDeclaredUidIdentifier('class');

        return decorators.map(function (dec) {
            return dec.expression;
        }).reverse().reduce(function (acc, decorator) {
            return buildClassDecorator({
                CLASS_REF: name,
                DECORATOR: decorator,
                INNER: acc
            }).expression;
        }, classPath.node);
    }

    /**
     * Given a class expression with method-level decorators, create a new expression
     * with the proper decorated behavior.
     */
    function applyMethodDecorators(path) {
        var hasMethodDecorators = path.node.body.body.some(function (node) {
            return (node.decorators || []).length > 0;
        });

        if (!hasMethodDecorators) return;

        return applyTargetDecorators(path, path.node.body.body);
    }

    /**
     * Given an object expression with property decorators, create a new expression
     * with the proper decorated behavior.
     */
    function applyObjectDecorators(path) {
        var hasMethodDecorators = path.node.properties.some(function (node) {
            return (node.decorators || []).length > 0;
        });

        if (!hasMethodDecorators) return;

        return applyTargetDecorators(path, path.node.properties);
    }

    /**
     * A helper to pull out property decorators into a sequence expression.
     */
    function applyTargetDecorators(path, decoratedProps) {
        var descName = path.scope.generateDeclaredUidIdentifier('desc');
        var valueTemp = path.scope.generateDeclaredUidIdentifier('value');

        var name = path.scope.generateDeclaredUidIdentifier(path.isClass() ? 'class' : 'obj');

        var exprs = decoratedProps.reduce(function (acc, node) {
            var decorators = node.decorators || [];
            node.decorators = null;

            if (decorators.length === 0) return acc;

            if (t.isClassProperty(node, { static: true })) {
                throw path.buildCodeFrameError('Static class property decorators are not yet supported.');
            }

            if (node.computed) {
                throw path.buildCodeFrameError('Computed method/property decorators are not yet supported.');
            }

            var property = t.isLiteral(node.key) ? node.key : t.stringLiteral(node.key.name);

            var target = path.isClass() && !node.static ? buildClassPrototype({
                CLASS_REF: name
            }).expression : name;

            var buildDecoratorCall = function buildDecoratorCall(descriptor, expr) {
                return buildPropertyDecorator({
                    TARGET: target,
                    DECORATOR: expr,
                    DESC: descName,
                    INNER: descriptor,
                    PROPERTY: property
                }).expression;
            };
            var buildDescriptorCreate = undefined;
            var buildDescriptorStore = undefined;

            if (t.isObjectProperty(node)) {
                // Read value from object and store in temp to be read from initializer.
                buildDescriptorCreate = function () {
                    return buildGetObjectInitializer({
                        TEMP: path.scope.generateDeclaredUidIdentifier('init'),
                        TARGET: target,
                        PROPERTY: property
                    }).expression;
                };

                // Trigger initializer and set descriptor.
                buildDescriptorStore = function () {
                    return buildSetObjectInitializer({
                        TARGET: target,
                        PROPERTY: property,
                        DESC: descName
                    }).expression;
                };
            } else if (t.isClassProperty(node)) {
                (function () {
                    var init = path.scope.generateDeclaredUidIdentifier('init');
                    var oldInitializer = undefined;
                    if (node.value) {
                        // Move the initializer to a temp variable and reassign to new temp.
                        oldInitializer = node.value;
                        acc.push(t.assignmentExpression('=', init, t.functionExpression(null, [], t.blockStatement([t.returnStatement(oldInitializer)]))));
                    }
                    node.value = buildClassInitializerCall({
                        INIT: init
                    }).expression;

                    // Create initializer that reads from new temp.
                    buildDescriptorCreate = function () {
                        return buildGetClassInitializer({
                            INIT: oldInitializer ? init : t.nullLiteral()
                        }).expression;
                    };

                    // Trigger initializer and set descriptor.
                    buildDescriptorStore = function () {
                        return buildSetClassInitializer({
                            INIT: init,
                            DESC: descName
                        }).expression;
                    };
                })();
            } else {
                // Read descriptor normally.
                buildDescriptorCreate = function () {
                    return buildGetDescriptor({
                        TARGET: target,
                        PROPERTY: property
                    }).expression;
                };

                // Store descriptor normally.
                buildDescriptorStore = function () {
                    return buildSetDescriptor({
                        TARGET: target,
                        PROPERTY: property,
                        DESC: descName
                    }).expression;
                };
            }

            return acc.concat(decorators.map(function (dec) {
                return dec.expression;
            }).reverse().reduce(buildDecoratorCall, t.assignmentExpression('=', descName, buildDescriptorCreate())), buildDescriptorStore());
        }, []);

        return t.sequenceExpression([t.assignmentExpression('=', name, path.node), t.sequenceExpression(exprs), name]);
    }

    return {
        inherits: require("babel-plugin-syntax-decorators"),

        visitor: {
            ExportDefaultDeclaration: function ExportDefaultDeclaration(path) {
                if (!path.get("declaration").isClassDeclaration()) return;

                var node = path.node;

                var ref = node.declaration.id || path.scope.generateUidIdentifier("default");
                node.declaration.id = ref;

                // Split the class declaration and the export into two separate statements.
                path.replaceWith(node.declaration);
                path.insertAfter(t.exportNamedDeclaration(null, [t.exportSpecifier(ref, t.identifier('default'))]));
            },
            ClassDeclaration: function ClassDeclaration(path) {
                var node = path.node;

                var ref = node.id || path.scope.generateUidIdentifier("class");

                path.replaceWith(t.variableDeclaration("let", [t.variableDeclarator(ref, t.toExpression(node))]));
            },
            ClassExpression: function ClassExpression(path) {
                // Create a replacement for the class node if there is one. We do one pass to replace classes with
                // class decorators, and a second pass to process method decorators.
                var decoratedClass = applyEnsureOrdering(path) || applyClassDecorators(path) || applyMethodDecorators(path);

                if (decoratedClass) path.replaceWith(decoratedClass);
            },
            ObjectExpression: function ObjectExpression(path) {
                var decoratedObject = applyEnsureOrdering(path) || applyObjectDecorators(path);

                if (decoratedObject) path.replaceWith(decoratedObject);
            }
        }
    };
};

var _babelTemplate = require('babel-template');

var _babelTemplate2 = _interopRequireDefault(_babelTemplate);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var buildClassDecorator = (0, _babelTemplate2.default)('\n  DECORATOR(CLASS_REF = INNER) || CLASS_REF;\n');

var buildPropertyDecorator = (0, _babelTemplate2.default)('\n  DESC = DECORATOR(TARGET, PROPERTY, INNER) || DESC;\n');

var buildClassPrototype = (0, _babelTemplate2.default)('\n  CLASS_REF.prototype;\n');

var buildGetDescriptor = (0, _babelTemplate2.default)('\n    Object.getOwnPropertyDescriptor(TARGET, PROPERTY);\n');

var buildSetDescriptor = (0, _babelTemplate2.default)('\n    DESC ? Object.defineProperty(TARGET, PROPERTY, DESC) : void 0;\n');

var buildGetObjectInitializer = (0, _babelTemplate2.default)('\n    (TEMP = Object.getOwnPropertyDescriptor(TARGET, PROPERTY).value, {\n        enumerable: true,\n        configurable: true,\n        writable: true,\n        initializer: function(){\n            return TEMP;\n        }\n    })\n');

var buildSetObjectInitializer = (0, _babelTemplate2.default)('\n    (DESC.value = DESC.initializer ? DESC.initializer.call(TARGET) : void 0) &&\n        Object.defineProperty(TARGET, PROPERTY, DESC);\n');

var buildGetClassInitializer = (0, _babelTemplate2.default)('\n    ({\n        enumerable: true,\n        configurable: true,\n        writable: true,\n        initializer: INIT,\n    })\n');

var buildSetClassInitializer = (0, _babelTemplate2.default)('\n    INIT = DESC.initializer;\n');

var buildClassInitializerCall = (0, _babelTemplate2.default)('\n    INIT ? INIT.apply(this) : void 0;\n');

;